@@ -22,6 +22,7 @@
 #define _DISK_H
 
 #include <sys/ioctl.h>
+#include <stdint.h>
 
 /* Snagged from linux/include/asm-ia64/ioctl.h */
 #define _IOC_NRBITS     8
@@ -63,21 +64,29 @@
 enum _bus_type {bus_type_unknown, isa, pci};
 enum _interface_type {interface_type_unknown,
 		      ata, atapi, scsi, usb,
-		      i1394, fibre, i2o, md};
+		      i1394, fibre, i2o, md,
+		      virtblk};
 
 
 unsigned int lcm(unsigned int x, unsigned int y);
 
 int disk_get_pci(int fd,
+		 int *interface_type,
 		 unsigned char *bus,
 		 unsigned char *device,
 		 unsigned char *function);
 
-int disk_info_from_fd(int fd, 
-		      int *interface_type,
-		      unsigned int *controllernum, 
-		      unsigned int *disknum,
-		      unsigned char *part);
+struct disk_info {
+	int interface_type;
+	unsigned int controllernum;
+	unsigned int disknum;
+	unsigned char part;
+	uint64_t major;
+	unsigned char minor;
+};
+
+int disk_info_from_fd(int fd, struct disk_info *info);
+
 
 int disk_get_partition_info (int fd, 
 			     uint32_t num,
