@@ -21,11 +21,12 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
-#include <stdint.h>
+#include <inttypes.h>
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <fcntl.h>
 #include <unistd.h>
+#include <errno.h>
 #include "disk.h"
 #include "scsi_ioctls.h"
 #include "gpt.h"
@@ -33,107 +34,183 @@
 
 #define BLKSSZGET  _IO(0x12,104)	/* get block device sector size */
 
+/* The major device number for virtio-blk disks is decided on module load time.
+ */
+static int
+get_virtblk_major(void)
+{
+	static int cached;
+	FILE *f;
+	char line[256];
+
+	if (cached != 0) {
+		return cached;
+	}
+
+	cached = -1;
+	f = fopen("/proc/devices", "r");
+	if (f == NULL) {
+		fprintf(stderr, "%s: opening /proc/devices: %s\n", __func__,
+			strerror(errno));
+		return cached;
+	}
+	while (fgets(line, sizeof line, f) != NULL) {
+		size_t len = strlen(line);
+		int major, scanned;
+
+		if (len == 0 || line[len - 1] != '\n') {
+			break;
+		}
+		if (sscanf(line, "%d %n", &major, &scanned) == 1 &&
+		    strcmp(line + scanned, "virtblk\n") == 0) {
+			cached = major;
+			break;
+		}
+	}
+	fclose(f);
+	if (cached == -1) {
+		fprintf(stderr, "%s: virtio-blk driver unavailable\n",
+			__func__);
+	}
+	return cached;
+}
+
 int
-disk_info_from_fd(int fd, 
-		  int *interface_type,
-		  unsigned int *controllernum, 
-		  unsigned int *disknum,
-		  unsigned char *part)
+disk_info_from_fd(int fd, struct disk_info *info)
 {
 	struct stat buf;
 	int rc;
-	uint64_t major;
-	unsigned char minor;
+
+	memset(info, 0, sizeof *info);
 	memset(&buf, 0, sizeof(struct stat));
 	rc = fstat(fd, &buf);
 	if (rc == -1) {
 		perror("stat");
 		return 1;
 	}
-	if (!(S_ISBLK(buf.st_mode) || S_ISREG(buf.st_mode))) {
+	if (S_ISBLK(buf.st_mode)) {
+		info->major = buf.st_rdev >> 8;
+		info->minor = buf.st_rdev & 0xFF;
+	}
+	else if (S_ISREG(buf.st_mode)) {
+		info->major = buf.st_dev >> 8;
+		info->minor = buf.st_dev & 0xFF;
+	}
+	else {
 		printf("Cannot stat non-block or non-regular file\n");
 		return 1;
 	}
-	major = buf.st_dev >> 8;
-	minor = buf.st_dev & 0xFF;
 
 	/* IDE disks can have up to 64 partitions, or 6 bits worth,
 	 * and have one bit for the disk number.
 	 * This leaves an extra bit at the top.
 	 */
-	if (major == 3) {
-		*disknum = (minor >> 6) & 1;
-		*controllernum = (major - 3 + 0) + *disknum;
-		*interface_type = ata;
-		*part    = minor & 0x3F;
+	if (info->major == 3) {
+		info->disknum = (info->minor >> 6) & 1;
+		info->controllernum = (info->major - 3 + 0) + info->disknum;
+		info->interface_type = ata;
+		info->part    = info->minor & 0x3F;
 		return 0;
 	}
-	else if (major == 22) {
-		*disknum = (minor >> 6) & 1;
-		*controllernum = (major - 22 + 2) + *disknum;
-		*interface_type = ata;
-		*part    = minor & 0x3F;
+	else if (info->major == 22) {
+		info->disknum = (info->minor >> 6) & 1;
+		info->controllernum = (info->major - 22 + 2) + info->disknum;
+		info->interface_type = ata;
+		info->part    = info->minor & 0x3F;
 		return 0;
 	}
-	else if (major >= 33 && major <= 34) {
-		*disknum = (minor >> 6) & 1;
-		*controllernum = (major - 33 + 4) + *disknum;
-		*interface_type = ata;
-		*part    = minor & 0x3F;
+	else if (info->major >= 33 && info->major <= 34) {
+		info->disknum = (info->minor >> 6) & 1;
+		info->controllernum = (info->major - 33 + 4) + info->disknum;
+		info->interface_type = ata;
+		info->part    = info->minor & 0x3F;
 		return 0;
 	}
-	else if (major >= 56 && major <= 57) {
-		*disknum = (minor >> 6) & 1;
-		*controllernum = (major - 56 + 8) + *disknum;
-		*interface_type = ata;
-		*part    = minor & 0x3F;
+	else if (info->major >= 56 && info->major <= 57) {
+		info->disknum = (info->minor >> 6) & 1;
+		info->controllernum = (info->major - 56 + 8) + info->disknum;
+		info->interface_type = ata;
+		info->part    = info->minor & 0x3F;
 		return 0;
 	}
-	else if (major >= 88 && major <= 91) {
-		*disknum = (minor >> 6) & 1;
-		*controllernum = (major - 88 + 12) + *disknum;
-		*interface_type = ata;
-		*part    = minor & 0x3F;
+	else if (info->major >= 88 && info->major <= 91) {
+		info->disknum = (info->minor >> 6) & 1;
+		info->controllernum = (info->major - 88 + 12) + info->disknum;
+		info->interface_type = ata;
+		info->part    = info->minor & 0x3F;
 		return 0;
 	}
  	
         /* I2O disks can have up to 16 partitions, or 4 bits worth. */
-	if (major >= 80 && major <= 87) {
-		*interface_type = i2o;
-		*disknum = 16*(major-80) + (minor >> 4);
-		*part    = (minor & 0xF);
+	if (info->major >= 80 && info->major <= 87) {
+		info->interface_type = i2o;
+		info->disknum = 16*(info->major-80) + (info->minor >> 4);
+		info->part    = (info->minor & 0xF);
 		return 0;
 	}
 
 	/* SCSI disks can have up to 16 partitions, or 4 bits worth
 	 * and have one bit for the disk number.
 	 */
-	if (major == 8) {
-		*interface_type = scsi;
-		*disknum = (minor >> 4);
-		*part    = (minor & 0xF);
+	if (info->major == 8) {
+		info->interface_type = scsi;
+		info->disknum = (info->minor >> 4);
+		info->part    = (info->minor & 0xF);
 		return 0;
 	}
-	else  if ( major >= 65 && major <= 71) {
-		*interface_type = scsi;
-		*disknum = 16*(major-64) + (minor >> 4);
-		*part    = (minor & 0xF);
+	else  if ( info->major >= 65 && info->major <= 71) {
+		info->interface_type = scsi;
+		info->disknum = 16*(info->major-64) + (info->minor >> 4);
+		info->part    = (info->minor & 0xF);
 		return 0;
 	}
-	    
+
+	if (get_virtblk_major() != -1 && get_virtblk_major() == info->major) {
+		info->interface_type = virtblk;
+		info->disknum = info->minor >> 4;
+		info->part = info->minor & 0xF;
+		return 0;
+	}
+
 	printf("Unknown interface type.\n");
 	return 1;
 }
 
+static int
+disk_get_virt_pci(const struct disk_info *info, unsigned char *bus,
+		  unsigned char *device, unsigned char *function)
+{
+	char inbuf[32], outbuf[128];
+	ssize_t lnksz;
+
+	if (snprintf(inbuf, sizeof inbuf, "/sys/dev/block/%" PRIu64 ":%u",
+		     info->major, info->minor) >= sizeof inbuf) {
+		return 1;
+	}
+
+	lnksz = readlink(inbuf, outbuf, sizeof outbuf);
+	if (lnksz == -1 || lnksz == sizeof outbuf) {
+		return 1;
+	}
+
+	outbuf[lnksz] = '\0';
+	if (sscanf(outbuf, "../../devices/pci0000:00/0000:%hhx:%hhx.%hhx",
+		   bus, device, function) != 3) {
+		return 1;
+	}
+	return 0;
+}
+
 static int
 disk_get_scsi_pci(int fd, 
+	     const struct disk_info *info,
 	     unsigned char *bus,
 	     unsigned char *device,
 	     unsigned char *function)
 {
 	int rc, usefd=fd;
 	struct stat buf;
-	char slot_name[8];
+	char slot_name[SLOT_NAME_SIZE];
 	unsigned int b=0,d=0,f=0;
 	memset(&buf, 0, sizeof(buf));
 	rc = fstat(fd, &buf);
@@ -151,11 +228,14 @@ disk_get_scsi_pci(int fd,
 		return 1;
 	}
 
-	rc = get_scsi_pci(usefd, slot_name);
+	rc = get_scsi_pci(usefd, slot_name, sizeof slot_name);
 	if (rc) {
 		perror("get_scsi_pci");
 		return rc;
 	}
+	if (strncmp(slot_name, "virtio", 6) == 0) {
+		return disk_get_virt_pci(info, bus, device, function);
+	}
 	rc = sscanf(slot_name, "%x:%x.%x", &b,&d,&f);
 	if (rc != 3) {
 		printf("sscanf failed\n");
@@ -189,19 +269,17 @@ disk_get_ide_pci(int fd,
 	     unsigned char *function)
 {
 	int num_scanned, procfd;
-	unsigned int b=0,d=0,disknum=0, controllernum=0;
-	unsigned char part=0;
+	struct disk_info info;
+	unsigned int b=0, d=0;
 	char procname[80], infoline[80];
-	size_t read_count;
-	int interface_type;
+	size_t read_count __attribute__((unused));
 	int rc;
 	
-	rc = disk_info_from_fd(fd, &interface_type, &controllernum,
-			       &disknum, &part);
+	rc = disk_info_from_fd(fd, &info);
 	if (rc) return rc;
 
 
-	sprintf(procname, "/proc/ide/ide%d/config", controllernum);
+	sprintf(procname, "/proc/ide/ide%d/config", info.controllernum);
 	
 	procfd = open(procname, O_RDONLY);
 	if (!procfd) {
@@ -223,43 +301,31 @@ disk_get_ide_pci(int fd,
 	return 0;
 }
 
-
-
-#if 0
-/* this is a list of devices */
-static int
-disk_get_md_parts(int fd)
-{
-	return 0;
-}
-#endif
-
 int
 disk_get_pci(int fd,
+	     int *interface_type,
 	     unsigned char *bus,
 	     unsigned char *device,
 	     unsigned char *function)
 {
-	int interface_type=interface_type_unknown;
-	unsigned int controllernum=0, disknum=0;
-	unsigned char part=0;
-	
-	disk_info_from_fd(fd,
-			  &interface_type,
-			  &controllernum, 
-			  &disknum,
-			  &part);
-	switch (interface_type) {
+	struct disk_info info;
+
+	disk_info_from_fd(fd, &info);
+	*interface_type = info.interface_type;
+
+	switch (info.interface_type) {
 	case ata:
 		return disk_get_ide_pci(fd, bus, device, function);
 		break;
 	case scsi:
-		return disk_get_scsi_pci(fd, bus, device, function);
+		return disk_get_scsi_pci(fd, &info, bus, device, function);
 		break;
 	case i2o:
 		break;
 	case md:
 		break;
+	case virtblk:
+		return disk_get_virt_pci(&info, bus, device, function);
 	default:
 		break;
 	}
@@ -352,7 +418,7 @@ msdos_disk_get_partition_info (int fd, legacy_mbr *mbr,
 		printf("******************************************************\n\n");
 		
 	}
-	else if (opts.write_signature) {
+	else if (!mbr->unique_mbr_signature && opts.write_signature) {
 		
 		/* MBR Signatures must be unique for the 
 		   EFI Boot Manager
@@ -375,8 +441,7 @@ msdos_disk_get_partition_info (int fd, legacy_mbr *mbr,
 			
 		/* Write it to the disk */
 		lseek(fd, 0, SEEK_SET);
-		write(fd, mbr, sizeof(*mbr));
-
+		rc = write(fd, mbr, sizeof(*mbr));
 	}
 	*(uint32_t *)signature = mbr->unique_mbr_signature;
 		
@@ -465,7 +530,7 @@ disk_get_partition_info (int fd,
 	legacy_mbr *mbr;
 	void *mbr_sector;
 	size_t mbr_size;
-	off_t offset;
+	off_t offset __attribute__((unused));
 	int this_bytes_read = 0;
 	int gpt_invalid=0, mbr_invalid=0;
 	int rc=0;
@@ -510,11 +575,11 @@ disk_get_partition_info (int fd,
 int
 main (int argc, char *argv[])
 {
-	int fd, rc;
+	int fd, rc, interface_type;
 	unsigned char bus=0,device=0,func=0;
 	if (argc <= 1) return 1;
 	fd = open(argv[1], O_RDONLY|O_DIRECT);
-	rc = disk_get_pci(fd, &bus, &device, &func);
+	rc = disk_get_pci(fd, &interface_type, &bus, &device, &func);
 	if (!rc) {
 		printf("PCI %02x:%02x.%02x\n", bus, device, func);
 	}
